# required: python3, pip
# install packages with:
# `pip install numpy matplotlib scipy tqdm`... etc
# see article for info installing gpufit/cpufit

import numpy as np
import numpy.random
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable, axes_size
import cycler
import scipy.optimize
import itertools
import warnings
import tqdm
import tqdm.contrib
import concurrent.futures
import psutil
import os
import sys
import pygpufit.gpufit as gf


def set_style():
    """sets the Defective Club matplotlib stylesheet"""
    rcparams_dict = {
        "axes.prop_cycle": cycler.cycler(
            color=[
                "#7239b3ff",
                "#008fd5",
                "#fc4f30",
                "#e5ae38",
                "#6d904f",
                "#8b8b8b",
            ]
        ),
        "lines.linewidth": 2,
        "lines.markersize": 3,
        "xtick.labelsize": 10,
        "xtick.major.size": 4,
        "xtick.direction": "in",
        "xtick.top": True,
        "ytick.labelsize": 10,
        "ytick.direction": "in",
        "ytick.major.size": 4,
        "ytick.right": True,
        "legend.fontsize": "small",
        "legend.loc": "lower left",
        "font.size": 8.0,
        "font.family": "monospace",
        "image.cmap": "plasma",
    }

    for optn, val in rcparams_dict.items():
        if isinstance(val, (list, tuple)):
            val = tuple(val)
        try:
            mpl.rcParams[optn] = val
        except KeyError:
            warn(
                f"mpl rcparams key '{optn}' not recognised as a valid rc parameter."
            )


def _add_colorbar(im, fig, ax, aspect=20, pad_fraction=1, **kwargs):
    """helped function, adds colorbar to 'im' returned by 'imshow'"""
    divider = make_axes_locatable(ax)
    width = axes_size.AxesY(ax, aspect=1.0 / aspect)
    pad = axes_size.Fraction(pad_fraction, width)
    cax = divider.append_axes("right", size=width, pad=pad)
    cbar = fig.colorbar(im, cax=cax, **kwargs)
    tick_locator = mpl.ticker.MaxNLocator(nbins=5)
    cbar.locator = tick_locator
    cbar.update_ticks()
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.linewidth = 0.5
    return cbar


def bounds_semicircle(pos, centre, radius, lw=2, right=True):
    """semicircle (+x or -x depending on 'right') shaped feature.
    (only the boundary of the feature, at linewidth 'lw')"""
    return (np.abs(radius - np.linalg.norm(pos - centre)) < lw) and (
        ((pos[1] - centre[1] > 0) and right)
        or ((pos[1] - centre[1] < 0) and not right)
    )


def get_freqs(num_freqs, freq_span):
    """return mw frequencies we 'sample'"""
    return np.linspace(2870 - freq_span // 2, 2870 + freq_span // 2, num_freqs)


def get_data_cube(
    heights,
    widths,
    bg_zeeman,
    ft_zeeman,
    ft_width_dif,
    ft_height_dif,
    cam_shape,
    ft_centre,
    ft_rad,
    ft_linewidth,
    num_freqs,
    freq_span,
    noise_sigma,
):
    """
    generate a 2-peak lorentzian data cube:
    two cocentric semicircles of opposite magnetic field"""
    heights = np.array(heights)
    widths = np.array(widths)
    raw_cube = np.empty((num_freqs, *cam_shape))
    freqs = get_freqs(num_freqs, freq_span)

    for ypos in range(cam_shape[0]):
        for xpos in range(cam_shape[1]):
            pos = np.array([ypos, xpos])
            heights_res = heights + ft_height_dif
            widths_res = widths + ft_width_dif
            if bounds_semicircle(
                pos, ft_centre, ft_rad, lw=ft_linewidth, right=True
            ):
                l_freq = 2870 - bg_zeeman // 2 - ft_zeeman // 2
                r_freq = 2870 + bg_zeeman // 2 + ft_zeeman // 2
            elif bounds_semicircle(
                pos, ft_centre, ft_rad, lw=ft_linewidth, right=False
            ):
                l_freq = 2870 - bg_zeeman // 2 + ft_zeeman // 2
                r_freq = 2870 + bg_zeeman // 2 - ft_zeeman // 2
            else:
                l_freq = 2870 - bg_zeeman // 2
                r_freq = 2870 + bg_zeeman // 2
                heights_res = heights
                widths_res = widths

            raw_cube[:, ypos, xpos] = (
                lorentz(freqs, [widths[0], l_freq, heights[0]])
                + lorentz(freqs, [widths[1], r_freq, heights[1]])
                + 1.0
            )

    rng = numpy.random.default_rng(12345)  # fix generator for reproducibility
    # add gaussian noise
    data_cube = raw_cube + rng.normal(
        0.0, noise_sigma, size=np.shape(raw_cube)
    )

    return data_cube


def lorentz(x, params):
    """lorentzian defined by fwhm, peak position, height"""
    w, x0, A = params
    R = ((x - x0) / (w / 2)) ** 2  # w/2 to convert from fwhm to hwhm
    return A * (1 / (1 + R))


def model(params, x):
    """the physics-informed model we want to fit the data to"""
    return 1.0 + lorentz(x, params[:3]) + lorentz(x, params[3:])


def resid(params, xs, pl):
    """residual: difference between measured data 'pl' and model"""
    return pl - model(params, xs)


def scipyfit_avg(freqs, data_cube, guess, npts=1000):
    """fit the average spectrum with scipy"""
    avg_spectrum = np.nanmean(data_cube, axis=(1, 2))
    fit_params = scipy.optimize.least_squares(
        resid, guess, args=(freqs, avg_spectrum)
    ).x
    fit_freqs = np.linspace(min(freqs), max(freqs), npts)
    return fit_freqs, model(fit_params, fit_freqs)


def cpufit_avg(freqs, data_cube, guess, npts=1000):
    """fit the average spectrum with cpufit"""
    # guess is [c, m, w_0, x_0, a_0, ...]
    avg_spectrum = np.nanmean(data_cube, axis=(1, 2))
    avg_params, state, _, _, _ = gf.fit(
        avg_spectrum[None, :].astype(np.float32),
        None,
        gf.ModelID.LORENTZ8_LINEAR,
        np.array(guess)[None, :].astype(np.float32),
        tolerance=1e-9,
        max_number_iterations=50,
        user_info=freqs.astype(np.float32),
        parameters_to_fit=np.array(
            [1 if i < 8 else 0 for i, _ in enumerate(guess)], np.int32
        ),
        platform="cpu",
    )
    fit_freqs = np.linspace(min(freqs), max(freqs), npts)

    def close_model(p, x):
        return (
            p[0, 0]
            + x * p[0, 1]
            + lorentz(x, p[0, 2:5])
            + lorentz(x, p[0, 5:8])
        )

    return fit_freqs, close_model(avg_params, fit_freqs)


def plot_avg_fits(freqs, avg_spectrum, fit_freqs, fit_res_list):

    fig_odmr, ax_odmr = plt.subplots(figsize=(4, 2))
    ax_odmr.plot(freqs, avg_spectrum, "o", ms=3.5, label="data", mfc="w")
    [
        ax_odmr.plot(fit_freqs, fit_res, "-", lw=1.5)
        for fit_res in fit_res_list
    ]

    ax_odmr.set_xlabel("RF frequency (MHz)")
    ax_odmr.set_ylabel("PL (a.u.)")

    fig_odmr.subplots_adjust(left=0.2, right=0.8, bottom=0.2, top=0.8)
    pass


def fit_wrapper(resid, guess, freqs, locator, fit_optns):
    """return the location of the fit, alongside the result"""
    fitres = scipy.optimize.least_squares(
        resid, guess, args=(freqs, locator[2]), **fit_optns
    )
    return (tuple(locator[:2]), fitres.x, fitres.jac)


def locator_generator(cube):
    """generates spectrum, while keeping track of pixel location (y,x)"""
    _, len_y, len_x = np.shape(cube)
    for y in range(len_y):
        for x in range(len_x):
            yield [y, x, cube[:, y, x]]


def get_images(results, image_shape, guess):
    """converts scipy wrapped results into images"""
    result_imgs = [np.full(image_shape, np.nan) for _ in enumerate(guess)]
    for pixel_loc, pixel_params, _ in results:
        for i, _ in enumerate(guess):
            result_imgs[i][pixel_loc] = pixel_params[i]
    return result_imgs


def serial_scipyfit_image(freqs, data_cube, guess):
    """fit the full data cube in series with scipy"""
    num_pixels = np.prod(data_cube.shape[1:3])
    series_results = list(
        tqdm.contrib.tmap(
            fit_wrapper,
            itertools.repeat(resid),
            itertools.repeat(guess),
            itertools.repeat(freqs),
            locator_generator(data_cube),
            itertools.repeat({"method": "trf"}),
            ascii=True,
            mininterval=1,
            total=num_pixels,
            unit="px",
            disable=False,
            desc="series",
        )
    )
    # series_results will be in shape: [(y, x), best_fit_params_array, jac_at_solution]
    return get_images(series_results, data_cube.shape[1:3], guess)


def parallel_scipyfit_image(freqs, data_cube, guess, chunksize, threads):
    """fit the full data cube in parallel with scipy"""
    num_pixels = np.prod(data_cube.shape[1:3])

    with concurrent.futures.ProcessPoolExecutor(
        max_workers=threads
    ) as executor:
        pool_results = list(
            tqdm.tqdm(
                executor.map(
                    fit_wrapper,
                    itertools.repeat(resid),
                    itertools.repeat(guess),
                    itertools.repeat(freqs),
                    locator_generator(data_cube),
                    itertools.repeat({"method": "trf"}),
                    chunksize=chunksize,
                ),
                ascii=True,
                mininterval=1,
                total=num_pixels,
                unit="px",
                disable=False,
                desc="pool  ",
            )
        )
    # pool_results will be in shape: [(y, x), best_fit_params_array, jac_at_solution]
    return get_images(pool_results, data_cube.shape[1:3], guess)


def gpufit_reshape(data_cube):
    """
    data_cube -> data_cube_reshaped, data_cube_positions, for gpufit/cpufit
    returns <reshaped_array>, <pixel position of each row in <reshaped_array>>
    """
    data_cube_reshaped = []  # (number_pixels, number_freqs)
    data_cube_positions = []  # pixel indices, matches shape of the above
    for y, x, spec in locator_generator(data_cube):
        data_cube_reshaped.append(spec)
        data_cube_positions.append(
            (y, x)
        )  # make sure this is a tuple for indexing!
    data_cube_positions = np.array(data_cube_positions, dtype=np.int32)
    data_cube_reshaped = np.array(data_cube_reshaped, dtype=np.float32)
    return data_cube_reshaped, data_cube_positions


def cpufit_image(freqs, data_cube, guess, bounds):
    """fit the full data cube in series with cpufit"""
    num_pixels = np.prod(data_cube.shape[1:3])
    guess_params = np.repeat([guess], repeats=num_pixels, axis=0).astype(
        np.float32
    )  # (number_pixels, number_params)
    constraint_types = np.array(
        [gf.ConstraintType.LOWER_UPPER for i in range(guess_params.shape[1])],
        dtype=np.int32,
    )
    constraints = np.tile(bounds, (num_pixels, 1)).astype(np.float32)

    data_cube_reshaped, data_cube_positions = gpufit_reshape(data_cube)

    cpufit_results, fit_states, _, _, execution_time = gf.fit_constrained(
        data_cube_reshaped,
        None,
        gf.ModelID.LORENTZ8_LINEAR,
        guess_params,
        tolerance=1e-12,
        max_number_iterations=50,
        # to compare with/without, set these two to None
        constraints=constraints,
        constraint_types=constraint_types,
        user_info=freqs.astype(np.float32),
        parameters_to_fit=np.array(
            [1 if i < 8 else 0 for i, _ in enumerate(guess)], np.int32
        ),  # yes this is required!
        platform="cpu",
    )
    print(f"cpufit execution time (s): {execution_time}", flush=True)
    # extract data from cpufit to same shape as fit_wrapper returns
    cpufit_params = []
    for params, pos in zip(cpufit_results, data_cube_positions):
        cpufit_params.append([tuple(pos), params, None])

    return get_images(cpufit_params, data_cube.shape[1:3], guess)


def get_bnv(fit_f0, fit_f1, bg_zeeman):
    gamma = 28  # MHz/T
    bnv = (fit_f1 - fit_f0) / (2 * gamma)
    bg = bg_zeeman / (2 * gamma)
    bnv -= bg
    return bnv


def get_avg_fwhm(fit_w0, fit_w1):
    return (fit_w0 + fit_w1) / 2.0


def plot_fit_images(bnv, avg_fwhm):
    fig, axs = plt.subplots(figsize=(6, 3), ncols=2)
    im_bnv = axs[0].imshow(bnv, cmap="bwr", vmin=-2, vmax=2)
    cbar = _add_colorbar(im_bnv, fig, axs[0])
    cbar.set_label("B_NV (T)", rotation=270, labelpad=5)

    im_w = axs[1].imshow(avg_fwhm, cmap="plasma", vmin=15, vmax=25)
    cbar = _add_colorbar(im_w, fig, axs[1])
    cbar.set_label("avg fwhm (MHz)", rotation=270, labelpad=10)

    for ax in axs:
        ax.set_xticks([])
        ax.set_yticks([])

    fig.subplots_adjust(wspace=0.5)
    return fig, axs

if __name__ == "__main__":
    set_style()

    cam_shape = (100, 100)
    num_freqs = 100
    freqs_span = 250  # MHz
    bg_zeeman = 100  # MHz zeeman split of background

    data_cube = get_data_cube(
        heights=(-0.012, -0.018),  # |-1>, |+1> resonance heights
        widths=(18, 22),
        bg_zeeman=bg_zeeman,
        ft_zeeman=50,  # marginal MHz split for feature
        ft_width_dif=5,  # add/sub width to/from feature resonance
        ft_height_dif=-0.004,  # add/sub height to/from feature resonance
        cam_shape=cam_shape,
        ft_centre=np.array([cam_shape[0] // 2, cam_shape[1] // 2]),
        ft_rad=cam_shape[0] // 4,
        ft_linewidth=4,
        # ft_linewidth=1,
        num_freqs=num_freqs,
        freq_span=freqs_span,
        noise_sigma=7.5e-3,  # gaussian noise
    )
    freqs = get_freqs(num_freqs, freqs_span)
    avg_spectrum = np.nanmean(data_cube, axis=(1, 2))

    # for concurrent.futures method
    chunksize = 10
    threads = 6

    scipy_guess = (20, 2820, -0.015, 20, 2920, -0.015)
    cpufit_guess = [
        1.0,  # c
        0.0,  # m
        20.0,  # width
        2820.0,  # freq
        -0.015,  # amplitude
        20.0,
        2920.0,
        -0.015,
    ]
    [cpufit_guess.append(0) for _ in range(6 * 3)]
    bounds = [
        0.9,  # c low
        1.1,  # c high
        -1e-3,  # m low
        1e-3,  # m high
        10,  # width low
        30,  # width high
        2790,  # ...
        2850,
        -0.1,
        -0.0001,
        10,
        30,
        2890,
        2950,
        -0.1,
        -0.0001,
    ]
    [bounds.append(np.nan) for _ in range(6 * 3 * 2)]

    # fit and plot average pl fits
    fit_freqs, sp_avg_fit = scipyfit_avg(freqs, data_cube, scipy_guess)
    _, cp_avg_fit = cpufit_avg(freqs, data_cube, cpufit_guess)
    # plot_avg_fits(freqs, avg_spectrum, fit_freqs, [sp_avg_fit])
    plot_avg_fits(freqs, avg_spectrum, fit_freqs, [sp_avg_fit, cp_avg_fit])

    # fit cube
    serial_res = serial_scipyfit_image(freqs, data_cube, scipy_guess)
    parallel_res = parallel_scipyfit_image(
        freqs, data_cube, scipy_guess, chunksize, threads
    )
    cpufit_res = cpufit_image(freqs, data_cube, cpufit_guess, bounds)
    
    # show results
    plot_fit_images(
        get_bnv(serial_res[1], serial_res[4], bg_zeeman),
        get_avg_fwhm(serial_res[0], serial_res[3]),
    )
    plot_fit_images(
        get_bnv(parallel_res[1], parallel_res[4], bg_zeeman),
        get_avg_fwhm(parallel_res[0], parallel_res[3]),
    )
    plot_fit_images(
        get_bnv(cpufit_res[2 + 1], cpufit_res[2 + 4], bg_zeeman),
        get_avg_fwhm(cpufit_res[2 + 0], cpufit_res[2 + 3]),
    )

    plt.show()
